﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Threading;
using System.Collections;

namespace WFRCS
{
    public class StatusChangedEventArgs : EventArgs
    {
        private string EventMsg;

        public string EventMessage
        {
            get
            {
                return EventMsg;
            }
            set
            {
                EventMsg = value;
            }
        }

        // Constructor for setting the event message
        public StatusChangedEventArgs(string strEventMsg)
        {
            EventMsg = strEventMsg;
        }
    }


    public delegate void StatusChangedEventHandler(object sender, StatusChangedEventArgs e);

    class Server
    {
        public static Hashtable htUsers;
        public static Hashtable htConnections;
        IPAddress ipAddress;
        TcpClient tcpClient;
        static StatusChangedEventArgs e;
        int _Port;
        Thread thrListener;
        TcpListener tlsClient;
        public static event StatusChangedEventHandler StatusChanged;
        bool ServRunning = false;

        public Server(IPAddress IP, int Port, int LimitUsers)
        {
            ipAddress = IP;
            _Port = Port;
            htUsers = new Hashtable(LimitUsers);
            htConnections = new Hashtable(LimitUsers);
        }

        public void Stop()
        {
            tcpClient.Close();
            tlsClient.Stop();
            thrListener.Abort();
        }

        public static void AddUser(TcpClient tcpUser, string strUsername)
        {
            Server.htUsers.Add(strUsername, tcpUser);
            Server.htConnections.Add(tcpUser, strUsername);
            SendAdminMessage(htConnections[tcpUser] + ":has joined us");            
        }

        public static void RemoveUser(TcpClient tcpUser)
        {
            if (htConnections[tcpUser] != null)
            {
                //SendAdminMessage(htConnections[tcpUser] + ":has left us");
                
                Server.htUsers.Remove(Server.htConnections[tcpUser]);
                Server.htConnections.Remove(tcpUser);
            }
        }

        public static void OnStatusChanged(StatusChangedEventArgs e)
        {
            StatusChangedEventHandler statusHandler = StatusChanged;
            if (statusHandler != null)
            {
                statusHandler(null, e);
            }
        }

        public static void SendAdminMessage(string Message)
        {
            StreamWriter swSenderSender;
            
            e = new StatusChangedEventArgs("Administrator:" + Message);
            OnStatusChanged(e);

            TcpClient[] tcpClients = new TcpClient[Server.htUsers.Count];
            Server.htUsers.Values.CopyTo(tcpClients, 0);
            for (int i = 0; i < tcpClients.Length; i++)
            {
                try
                {
                    if (Message.Trim() == "" || tcpClients[i] == null)
                    {
                        continue;
                    }
                    swSenderSender = new StreamWriter(tcpClients[i].GetStream());
                    swSenderSender.WriteLine("Administrator:" + Message);
                    swSenderSender.Flush();
                    swSenderSender = null;
                }
                catch 
                {
                    RemoveUser(tcpClients[i]);
                }
            }
        }

        public static void SendMessage(string From, string Message)
        {
            StreamWriter swSenderSender;

            e = new StatusChangedEventArgs(From + ":" + Message);
            OnStatusChanged(e);

            TcpClient[] tcpClients = new TcpClient[Server.htUsers.Count];
            Server.htUsers.Values.CopyTo(tcpClients, 0);
            for (int i = 0; i < tcpClients.Length; i++)
            {
                try
                {
                    if (Message.Trim() == "" || tcpClients[i] == null)
                    {
                        continue;
                    }
                    swSenderSender = new StreamWriter(tcpClients[i].GetStream());
                    swSenderSender.WriteLine(From + ":" + Message);
                    swSenderSender.Flush();
                    swSenderSender = null;
                }
                catch 
                {
                    RemoveUser(tcpClients[i]);
                }
            }
        }

        public void StartListening()
        {
            IPAddress ipaLocal = ipAddress;
            tlsClient = new TcpListener(ipAddress, _Port);
            tlsClient.Start();
            ServRunning = true;
            thrListener = new Thread(KeepListening);
            thrListener.Start();
        }

        private void KeepListening()
        {
            while (ServRunning == true)
            {
                tcpClient = tlsClient.AcceptTcpClient();
                Connection newConnection = new Connection(tcpClient);
            }
        }

        public bool Running { get { return ServRunning; } }

        //~Server()
        //{
        //    SendAdminMessage("Server is disconnected.");
        //    Thread.Sleep(3000);
        //    TcpClient[] tcpClients = new TcpClient[Server.htUsers.Count];
        //    Server.htUsers.Values.CopyTo(tcpClients, 0);
        //    for (int i = 0; i < tcpClients.Length; i++)
        //    {
        //        tcpClients[i].Close();
        //    }

        //    htConnections.Clear();
        //    htUsers.Clear();
        //    thrListener.Abort();
        //    tlsClient.Stop();
        //    ServRunning = false;
        //    ipAddress = null;
        //}
    }

    class Connection
    {
        TcpClient tcpClient;
        private Thread thrSender;
        private StreamReader srReceiver;
        private StreamWriter swSender;
        private string currUser;
        private string strResponse;

        public Connection(TcpClient tcpCon)
        {
            tcpClient = tcpCon;
            thrSender = new Thread(AcceptClient);
            thrSender.Start();
        }

        private void CloseConnection()
        {
            tcpClient.Close();
            srReceiver.Close();
            swSender.Close();
        }

        private void AcceptClient()
        {
            srReceiver = new System.IO.StreamReader(tcpClient.GetStream());
            swSender = new System.IO.StreamWriter(tcpClient.GetStream());

            currUser = srReceiver.ReadLine();

            if (currUser != "")
            {
                if (Server.htUsers.Contains(currUser) == true)
                {
                    swSender.WriteLine("0|This username already exists.");
                    swSender.Flush();
                    CloseConnection();
                    return;
                }
                else if (currUser.ToLower() == "administrator")
                {
                    swSender.WriteLine("0|This username is reserved.");
                    swSender.Flush();
                    CloseConnection();
                    return;
                }
                else
                {
                    swSender.WriteLine("1");
                    swSender.Flush();

                    Server.AddUser(tcpClient, currUser);
                }
            }
            else
            {
                CloseConnection();
                return;
            }

            try
            {
                while ((strResponse = srReceiver.ReadLine()) != "")
                {
                    if (strResponse == null)
                    {
                        Server.RemoveUser(tcpClient);
                    }
                    else
                    {
                        Server.SendMessage(currUser, strResponse);
                    }
                }
            }
            catch
            {
                Server.RemoveUser(tcpClient);
            }
        }
    }
}
